import numpy as np
from scipy.interpolate import interp1d
import matplotlib as mpl
from datetime import datetime, timezone
from matplotlib.colors import LinearSegmentedColormap
import warnings
import pyspedas
import logging
from copy import copy

def get_bin_boundaries(bin_centers:np.ndarray, ylog:bool = False):
    """ Calculate a list of bin boundaries from a 1-D array of bin center values.

    Parameters
    ----------
    bin_centers: np.ndarray
    Array of Y bin center values

    ylog: bool
    If True, compute the bin boundaries in log space

    Returns
    -------
    tuple
    bin_boundaries: np.ndarray[float]
        Floating point array of bin boundaries computed from the bin centers.  Output array will be one element longer than the input array.
    direction: int
        Flag increasing or decreasing bin order: +1 increasing, -1 decreasing, 0 indeterminate

    """

    nbins = len(bin_centers)
    # Bin boundaries need to be floating point, even if the original bin values are
    # integers.  Initialize to all nans. Since the outputs are boundaries, not centers,
    # there is an extra element in the output array.
    outbins = np.zeros(nbins+1,dtype=np.float64)
    outbins[:] = np.nan

    # If we're working in log space, do the transform before filtering for finite values.
    # THEMIS ESA can have 0.0 bin centers with log scaling!


    if ylog:
        # There might be bin centers equal to 0.0 (e.g. THEMIS ESA).  Replace them with half the next larger
        # bin center.  Any bin centers less than zero will get turned to NaNs when we take logs
        # (and the corresponding data bins effectively removed).
        clean_bins = copy(bin_centers)
        zero_idx = np.where(bin_centers == 0.0)
        if len(zero_idx[0]) > 0:
            clean_bins[zero_idx] = np.nan
            clean_bins[zero_idx] = np.nanmin(clean_bins)/2.0
        working_bins = np.log10(clean_bins)
    else:
        working_bins = bin_centers

    # Check for all nans, or only one finite value
    idx_finite = np.where(np.isfinite(working_bins))

    if type(idx_finite) is tuple:
        idx_finite = idx_finite[0]

    if len(idx_finite) == 0:
        # Return all nans, indeterminate direction
        return outbins, 0
    elif len(idx_finite) == 1:
        idx = idx_finite[0]
        # Only a single bin, so we have to make up some bin boundaries
        if ylog:
            outbins[idx] = bin_centers[idx]/2.0
            outbins[idx+1] = bin_centers[idx]*2.0
        else:
            outbins[idx] = bin_centers[idx] - 1.0
            outbins[idx+1] = bin_centers[idx] + 1.0
        logging.warning("get_bin_boundaries: only one finite bin detected at index %d", idx)
        logging.warning("bin center: %f   bin boundaries: [%f, %f]",bin_centers[idx],outbins[idx], outbins[idx+1])
        # Return boundaries around the single finite bin, direction is increasing
        return outbins, 1

    # The usual case: we have at least two finite bins
    # Are they in increasing or decreasing order?

    if working_bins[idx_finite[0]] > working_bins[idx_finite[-1]]:
        direction = -1
    elif working_bins[idx_finite[0]] < working_bins[idx_finite[-1]]:
        direction = 1
    else:
        # All finite bins are the same?  Or nonmonotonic?  I guess it could happen.
        direction = 0
        logging.warning("get_bin_boundaries: First and last finite bin values are identical, may be nonmonotonic or all the same?")


    # We need to make sure that no NaNs are sandwiched between finite values
    good_bin_count = len(idx_finite)
    leading_nan_count = idx_finite[0]
    trailing_nan_count = nbins - idx_finite[-1] - 1

    if good_bin_count + leading_nan_count + trailing_nan_count != nbins:
        logging.warning("get_bin_boundaries: may contain nans between finite values. Total bin count: %d,  leading nans: %d, trailing_nans: %d, finite vals: %d",
                        nbins, leading_nan_count, trailing_nan_count, good_bin_count)

    # Now compute bin boundaries from all the finite bin centers
    finite_bins = np.copy(working_bins[idx_finite])
    edge_count = good_bin_count+1
    goodbins = np.zeros(edge_count, dtype=np.float64)

    goodbins[0] = finite_bins[0] - (finite_bins[1] - finite_bins[0]) / 2.0
    goodbins[1:edge_count-1] = (finite_bins[:-1] + finite_bins[1:]) / 2.0
    goodbins[edge_count-1] = finite_bins[-1] + (finite_bins[-1] - finite_bins[-2]) / 2.0

    # Deal with any possible leading or trailing nans
    if leading_nan_count > 0:
        outbins[0:leading_nan_count] = np.nan
    outbins[leading_nan_count:leading_nan_count+edge_count] = goodbins[:]
    if trailing_nan_count > 0:
        outbins[leading_nan_count+edge_count:nbins+2] = np.nan

    # Go back to linear space
    if ylog:
        outbins = 10.0**outbins

    return outbins, direction



def specplot_make_1d_ybins(values: np.ndarray, vdata:np.ndarray, ylog:bool, min_ratio:float = 0.001, no_regrid=False):
    """ Convert 2-D Y-bin arrays of bin center values to a 1-D array and rebin the data array

    Parameters
    ----------
    values : np.ndarray
    A 2-D array of values to be plotted as a spectrogram

    vdata: np.ndarray
    A 1-d or 2-D array of values representing center values of Y axis bins.
    If 1-d, bin centers are constant, otherwise they are assumed to be time-varying.

    ylog: bool
    If True, compute the bin boundaries in log space

    min_ratio: float
    Specifies a threshold for determining whether adjacent bins should be combined in a
    "thinning" process.  The default value of .001 represents a 1-pixel difference in
    where the bin boundaries are rendered if the Y axis is 1000 pixels high.

    no_regrid: bool
    If True, skip rebinning the data array and only return the bin boundaries.

    Returns
    -------
    tuple
    regridded_zdata:np.ndarray
        The result of regridding the values array with the new, potentially differwnt bin boundaries
    bin_boundaries_1d:np.ndarray
        The new bin boundaries

    Notes
    -----

    We find the union of all the bin boundaries for time-varying bins, and use that
    instead of an arbitrary high-resolution grid as the resampling target y-values.
    Any NaN values found in the input bin centers (1D or 2D) are dealt with.

    Allows for monotonically increasing or decreasing bin values, and bins that change over time.
    2D bin center arrays with some times having ascending values and other times having descending values
    are allowed.

    """
    #logging.info("Starting 1D vbins processing")
    ntimes = values.shape[0]
    bins_1d = False

    # Determine bin boundaries at each time step (or for all time, with 1-d bin center arrays),
    # weeding out NaNs in the bin center values. Form the union of all the individual bin
    # boundary sets (keeping the time-specific boundary sets, to use when rebinning later
    # Also determine the direction of increase at each time step, and flag any time steps
    # where a direction cannot be determined.

    if len(vdata.shape) == 1:
        #logging.info("Starting 1D vbins boundary processing")
        bins_1d = True
        result = get_bin_boundaries(vdata, ylog=ylog)
        vdata_bins = result[0]
        vdata_direction = result[1]
        bin_boundaries_set = set(vdata_bins)
        input_bin_center_count = len(vdata)
    else:  # 2-d V
        #logging.info("Starting 2D vbins boundary processing")
        bins_1d = False
        input_bin_center_count = vdata.shape[1]
        vdata_bins = np.zeros((ntimes, input_bin_center_count + 1), dtype=np.float64)
        vdata_direction = np.zeros(ntimes, dtype=np.int64)

        # The sentinel values are inserted because arrays with nans will not compare equal,
        # even if the nans are in same places
        sentinel = 1e31
        result = get_bin_boundaries(vdata[0, :], ylog=ylog)
        vdata_bins[0,:] = result[0]
        vdatadir_thistime = result[1]
        vdata_direction[0] = vdatadir_thistime
        pbins = vdata_bins[0,:]
        bin_boundaries_set = set(pbins)
        p = np.copy(vdata[0,:])
        p[np.where(~np.isfinite(p))] = sentinel
        # Now that everything is initialized, go through each time index
        # and maintain a set of all the bin boundaries seen so far.
        for i in range(ntimes-1):
            k = i+1
            t=np.copy(vdata[k,:])
            t[np.where(~np.isfinite(t))] = sentinel
            diff=t-p
            if np.any(diff):
                # bin values have changed, recalculate boundaries and add to running set
                #print(k)
                result = get_bin_boundaries(vdata[k,:], ylog=ylog)
                u = result[0]
                vdatadir_thistime = result[1]
                pbins = u
                vdata_bins[k,:] = u
                vdata_direction[k] = vdatadir_thistime
                uset = set(u)
                bin_boundaries_set = bin_boundaries_set | uset
                p = t
            else:
                # bin values have not channged at this time step, use previously computed values
                vdata_bins[k,:] = pbins
                vdata_direction[k] = vdatadir_thistime

    if bins_1d:
        if vdata_direction == 1:
            #print("1d bins are increasing")
            pass
        elif vdata_direction == -1:
            #print("1d bins are decreasing")
            pass
        else:
            logging.warning("specplot_make_1d_ybins: Direction of increase of 1-D input bin centers are indeterminate (all-nan, all-same, or nonmonotonic)")
            pass
    else:
        inc_count = (vdata_direction == 1).sum()
        dec_count = (vdata_direction == -1).sum()
        ind_count = (vdata_direction == 0).sum()
        #print("2d bins increasing: ", str(inc_count))
        #print("2d bins decreasing: ", str(dec_count))
        #print("2d bins indeterminate: ", str(ind_count))
        if ind_count > 0:
            logging.warning("specplot_make_1d_ybins: Direction of increase of input bin centers was indeterminate (all-nan, all-same, or non-monotonic) at %d of %d time indices.", ind_count, ntimes)

    # Convert the bin boundary set back to an array
    #logging.info("Done finding bin boundaries, sorting")
    vdata_unsorted = np.array(list(bin_boundaries_set))

    # Clean nans and sentinel values (e.g. may be present in FAST y bin values)
    vdata_finite = [val for val in vdata_unsorted if np.isfinite(val)]

    # Sort in ascending order
    output_bin_boundaries = np.sort(vdata_finite)

    output_bin_boundary_len = len(output_bin_boundaries)

    # It is possible (e.g. ELFIN) that some bin boundaries are very close, but not equal
    # (less than a pixel high).  We might want to thin out any bin boundaries within
    # "epsilon" of the previous bin.

    ymax=output_bin_boundaries[output_bin_boundary_len-1]
    ymin=output_bin_boundaries[0]
    yrange = ymax-ymin
    #logging.info("Thinning bin boundaries")
    # With the default min_ratio, epsilon is about one pixel in the y direction for a typical plot size and dpi
    # If min_ratio is 0, the effect is that no bin boundaries will be discarded.

    if ylog:
        epsilon = (np.log10(ymax)-np.log10(ymin)) * min_ratio
    else:
        epsilon = (ymax-ymin)*min_ratio
    diff=output_bin_boundaries[1:]-output_bin_boundaries[:output_bin_boundary_len-1]

    # logging.info("Specplot 1-D y bins before thinning: array size %d, yrange %f, smallest difference %f, epsilon %f,  ratio %f",len(output_bin_boundaries),yrange,np.min(diff),epsilon, (ymax-ymin)/np.min(diff))

    last_val = output_bin_boundaries[0]
    thinned_list = [output_bin_boundaries[0]]
    for i in range(output_bin_boundary_len):
        val = output_bin_boundaries[i]
        if ylog:
            diff = np.log10(val) - np.log10(last_val)
        else:
            diff = val-last_val
        if abs(diff) > epsilon:
            thinned_list.append(val)
            last_val = val

    output_bin_boundaries = np.array(thinned_list)

    #logging.info("Done thinning bin boundaries")
    output_bin_boundary_len = len(output_bin_boundaries)
    # There could be NaNs (e.g. FAST)
    ymax=output_bin_boundaries[output_bin_boundary_len-1]
    ymin=output_bin_boundaries[0]
    yrange = ymax-ymin
    diff=output_bin_boundaries[1:]-output_bin_boundaries[:output_bin_boundary_len-1]

    # logging.info("Specplot 1-D y bins after thinning: array size %d, yrange %f, smallest difference %f, epsilon %f,  ratio %f",len(output_bin_boundaries),yrange,np.min(diff),epsilon, (ymax-ymin)/np.min(diff))

    if no_regrid:
        return output_bin_boundaries
    #logging.info("Started rebinning (new style)")
    if values.dtype.kind == 'f':
        fill = np.nan
    else:
        fill = 0

    # Now we rebin the input data array into the output array, using both the original
    # and combined bin boundaries.

    # The output value array should have a y dimension one less than the bin boundary count
    out_values = np.zeros((ntimes, output_bin_boundary_len - 1), dtype=values.dtype)
    out_values[:,:] = fill

    # Note that output_bin_boundaries is always monotonically increasing, but
    # vdata_bins (the original inputs) can be monotonically decreasing

    for time_index in range(ntimes):
        if len(vdata.shape) == 1:
            input_bin_boundaries = vdata_bins
            input_bin_centers = vdata
            direction = vdata_direction
        else:
            input_bin_boundaries = vdata_bins[time_index, :]
            input_bin_centers = vdata[time_index,:]
            direction = vdata_direction[time_index]

        #print("Time index: " + str(time_index))
        #print("Input bin boundaries:")
        #print(input_bin_boundaries)
        #print("Input bin centers")
        #print(input_bin_centers)
        #print("Output bin boundaries")
        #print(output_bin_boundaries)
        #print("Data values at time step:")
        #print(values[time_index,:])
        if direction == 1:
            # Increasing bin values
            lower_bound_indices = np.searchsorted(output_bin_boundaries, input_bin_boundaries[0:-1],side="left")
            upper_bound_indices = np.searchsorted(output_bin_boundaries, input_bin_boundaries[1:],side="left")
        elif direction == -1:
            lower_bound_indices = np.searchsorted(output_bin_boundaries, input_bin_boundaries[1:],side="left")
            upper_bound_indices = np.searchsorted(output_bin_boundaries, input_bin_boundaries[0:-1],side="left")
        else:
            continue
        #print("Lower bound indices:")
        #print(lower_bound_indices)
        #print("Upper bound indices")
        #print(upper_bound_indices)

        for i in range(input_bin_center_count):
            if not np.isfinite(input_bin_centers[i]):
                pass
            else:
                #print("input bin index: " + str(i))
                #print("input bin center: " + str(input_bin_centers[i]))
                #print("output lower bound index: " + str(lower_bound_indices[i]))
                #print("output upper bound index: " + str(upper_bound_indices[i]))
                #print("output bin value range: " + str(output_bin_boundaries[lower_bound_indices[i]:upper_bound_indices[i]+1]))
                out_values[time_index,lower_bound_indices[i]:upper_bound_indices[i]] = values[time_index, i]
                #print(out_values)
    #logging.info("Done making 1D Y bins")

    return out_values, output_bin_boundaries

def specplot(
    var_data,
    var_times,
    this_axis,
    yaxis_options,
    zaxis_options,
    plot_extras,
    colorbars,
    axis_font_size,
    fig,
    variable,
    time_idxs=None,
    style=None,
):
    """
    Plot a tplot variable as a spectrogram

    Parameters
    ----------
    var_data: dict
        A tplot dictionary containing the data to be plotted
    var_times: array of datetime objects
        An array of datetime objects specifying the time axis
    this_axis
        The matplotlib object for the panel currently being plotted
    yaxis_options: dict
        A dictionary containing the Y axis options to be used for this variable
    zaxis_options: dict
        A dictionary containing the Z axis (spectrogram values) options to be used for this variable
    plot_extras: dict
        A dictionary containing additional plot options for this variable
    colorbars: dict
        The data structure that contains information (for all variables) for creating colorbars.
    axis_font_size: int
        The font size in effect for this axis (used to create colorbars)
    fig: matplotlib.figure.Figure
        A matplotlib figure object to be used for this plot
    variable: str
        The name of the tplxxot variable to be plotted (used for log messages)
    time_idxs: array of int
        The indices of the subset of times to use for this plot. Defaults to None (plot all timestamps).
    style
        A matplotlib style object to be used for this plot. Defaults to None.

    Returns
    -------
        True
    """
    alpha = plot_extras.get("alpha")
    spec_options = {"shading": "auto", "alpha": alpha}
    ztitle = zaxis_options["axis_label"]

    zlog_str = zaxis_options["z_axis_type"]
    ylog_str = yaxis_options["y_axis_type"]
    # Convert zlog_str and ylog_str to bool
    ylog = False
    zlog = False
    if "log" in ylog_str.lower():
        ylog = True
    if "log" in zlog_str.lower():
        zlog = True
    #logging.info("ylog_str is " + str(ylog_str))
    #logging.info("zlog_str is " + str(zlog_str))

    yrange = yaxis_options["y_range"]
    if yrange[0] is None or not np.isfinite(yrange[0]):
        yrange[0] = None
    if yrange[1] is None or not np.isfinite(yrange[1]):
        yrange[1] = None

    if zaxis_options.get("z_range") is not None:
        zrange = zaxis_options["z_range"]
    else:
        zrange = [None, None]

    if zaxis_options.get("axis_subtitle") is not None:
        zsubtitle = zaxis_options["axis_subtitle"]
    else:
        zsubtitle = ""

    # Clean up any fill values in data array
    #set -1.e31 fill values to NaN, jmjm, 2024-02-29
    ytp = np.where(var_data.y==-1.e31,np.nan,var_data.y)
    var_data.y[:,:] = ytp[:,:]

    if zlog:
        zmin = np.nanmin(var_data.y)
        zmax = np.nanmax(var_data.y)
        # gracefully handle the case of all NaNs in the data, but log scale set
        # all 0 is also a problem, causes a crash later when creating the colorbar
        if np.isnan(var_data.y).all():
            # no need to set a log scale if all the data values are NaNs, or all zeroes
            spec_options["norm"] = None
            spec_options["vmin"] = zrange[0]
            spec_options["vmax"] = zrange[1]
            logging.info("Variable %s contains all-NaN data", variable)
        elif not np.any(var_data.y):
            # properly handle all 0s in the data
            spec_options["norm"] = None
            spec_options["vmin"] = zrange[0]
            spec_options["vmax"] = zrange[1]
            logging.info("Variable %s contains all-zero data", variable)
        else:
            spec_options["norm"] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])
    else:
        spec_options["norm"] = None
        spec_options["vmin"] = zrange[0]
        spec_options["vmax"] = zrange[1]

    cmap = None

    if plot_extras.get("colormap") is not None:
        cmap = plot_extras["colormap"][0]
    else:
        # default to the SPEDAS color map if the user doesn't have a MPL style set
        if style is None:
            cmap = "spedas"

    # kludge to add support for the 'spedas' color bar
    if cmap == "spedas":
        _colors = pyspedas.tplot_tools.spedas_colorbar
        spd_map = [
            (np.array([r, g, b])).astype(np.float64) / 256
            for r, g, b in zip(_colors.r, _colors.g, _colors.b)
        ]
        cmap = LinearSegmentedColormap.from_list("spedas", spd_map)

    spec_options["cmap"] = cmap

    input_zdata = var_data.y[time_idxs, :]
    input_times = var_data.times[time_idxs]

    # Figure out which attribute to use for Y bin centers
    #allow use of v1, v2, jmm, 2024-03-20
    if len(var_data) == 3:
        if hasattr(var_data,'v'):
            input_bin_centers = var_data.v
        elif hasattr(var_data,'v1'):
            input_bin_centers = var_data.v1
        else:
            logging.warning("Multidimensional variable %s has no v or v1 attribute",variable)
    elif len(var_data) == 4:
        if hasattr(var_data, 'v1'):
            if 'spec_dim_to_plot' in plot_extras:
                if plot_extras['spec_dim_to_plot'] == 'v1':
                    input_bin_centers = var_data.v1
        if hasattr(var_data, 'v2'):
            if 'spec_dim_to_plot' in plot_extras:
                if plot_extras['spec_dim_to_plot'] == 'v2':
                    input_bin_centers = var_data.v2
    else:
        logging.warning("Too many dimensions on the variable: " + variable)
        return

    # Clean up any fill values in bin center array
    vtp = np.where(input_bin_centers == -1.e31, np.nan, input_bin_centers)
    if len(vtp.shape) == 1:
        input_bin_centers[:] = vtp[:]
    else:
        input_bin_centers[:, :] = vtp[:, :]

    if len(input_bin_centers.shape) > 1:
        # time varying 'v', need to limit the values to those within the requested time range
        input_bin_centers = input_bin_centers[time_idxs, :]

    # This call flattens any time-varying bin boundaries into a 1-d list (out_vdata)
    # and regrids the data array to the new y bin count (regridded_zdata)

    #logging.info("Starting specplot processing")

    regridded_zdata, bin_boundaries_1d = specplot_make_1d_ybins(input_zdata, input_bin_centers, ylog)

    # At this point, bin_boundaries_1d, the array of bin boundaries, is guaranteed to be
    # 1-dimensional, in ascending order, with all finite values. It has one more element
    # than the Y dimension of regridded_zdata.  The Y dimension of regridded_zdata may have changed
    # as a result of flattening a 2-D out_vdata input.
    # If ylog==True, all values in bin_boundaries_1d will be strictly positive.

    assert(len(bin_boundaries_1d.shape) == 1) # bin boundaries are 1-D
    assert(len(regridded_zdata.shape) == 2) # output array is 2-D
    assert(bin_boundaries_1d.shape[0] == regridded_zdata.shape[1]+1) # bin boundaries have one more element in Y dimension
    assert(np.isfinite(bin_boundaries_1d.all())) # no nans in bin boundaries
    assert(bin_boundaries_1d[-1] > bin_boundaries_1d[0]) # bin boundaries in ascending order
    if ylog:
        assert(np.all(bin_boundaries_1d > 0.0)) # bin boundaries all positive if log scaling

    # Get min and max bin boundaries
    vmin = np.min(bin_boundaries_1d)
    vmax = np.max(bin_boundaries_1d)

    #could also have a fill in yrange
    #    if yrange[0] == -1e31: #This does not work sometimes?
    if yrange[0] < -0.9e31:
        yrange[0] = vmin
    if yrange[1] < -0.9e31:
        yrange[1] = vmax


    #logging.info("Starting specplot time boundary processing")
    input_unix_times = np.int64(input_times) / 1e9
    result = get_bin_boundaries(input_unix_times)
    # For pcolormesh, we also want bin boundaries (not center values) on the time axis
    time_boundaries_dbl = result[0]
    time_boundaries_ns = np.int64(time_boundaries_dbl*1e9)
    # Now back to numpy datetime64
    time_boundaries = np.array(time_boundaries_ns, dtype='datetime64[ns]')
    #logging.info("Done with specplot initial processing")
    # If the user set a yrange with the 'options' command, nothing is needed here
    # since tplot takes care of it.   If not, set it here to the min/max finite bin
    # boundaries.  If left unspecified, pcolormesh might do something weird to the y limits.
    if yaxis_options.get('y_range_user') is None:
        this_axis.set_ylim([vmin, vmax])


    # automatic interpolation options
    if yaxis_options.get("x_interp") is not None:
        x_interp = yaxis_options["x_interp"]

        # interpolate along the x-axis
        if x_interp:
            if yaxis_options.get("x_interp_points") is not None:
                nx = yaxis_options["x_interp_points"]
            else:
                fig_size = fig.get_size_inches() * fig.dpi
                nx = fig_size[0]

            if zlog:
                zdata = np.log10(regridded_zdata)
            else:
                zdata = regridded_zdata

            zdata[zdata < 0.0] = 0.0
            zdata[zdata == np.nan] = 0.0

            # convert to floats for the interpolation
            spec_unix_times = np.int64(var_data.times[time_idxs]) / 1e9

            # interpolate in the x-direction
            interp_func = interp1d(
                spec_unix_times, zdata, axis=0, bounds_error=False, kind="linear"
            )
            out_times = (
                np.arange(0, nx, dtype=np.float64)
                * (spec_unix_times[-1] - spec_unix_times[0])
                / (nx - 1)
                + spec_unix_times[0]
            )
            regridded_zdata = interp_func(out_times)

            if zlog:
                regridded_zdata = 10**regridded_zdata

            # Convert time bin centers to bin boundaries
            result = get_bin_boundaries(out_times, ylog=False)
            # Convert unix times back to np.datetime64[ns] objects
            unix_time_boundaries_int64 = np.int64(result[0]*1e9)
            time_boundaries = np.array(unix_time_boundaries_int64, dtype="datetime64[ns]")

    if yaxis_options.get("y_interp") is not None:
        y_interp = yaxis_options["y_interp"]

        if y_interp:
            if yaxis_options.get("y_interp_points") is not None:
                ny = yaxis_options["y_interp_points"]
            else:
                fig_size = fig.get_size_inches() * fig.dpi
                ny = fig_size[1]

            if zlog:
                zdata = np.log10(regridded_zdata)
            else:
                zdata = regridded_zdata

            if ylog:
                vdata = np.log10(bin_boundaries_1d)
                ycrange = np.log10(yrange)
            else:
                vdata = bin_boundaries_1d
                ycrange = yrange

            if not np.isfinite(ycrange[0]):
                ycrange = [np.min(vdata), yrange[1]]

            zdata[zdata < 0.0] = 0.0
            zdata[zdata == np.nan] = 0.0  # does not work

            # vdata was calculated from 1-D bin boundaries, not bin centers.
            # We need to go to bin centers for interpolation
            uppers = vdata[1:]
            lowers = vdata[0:-1]
            centers = (uppers+lowers)/2.0

            interp_func = interp1d(centers, zdata, axis=1, bounds_error=False)
            out_vdata_centers = (
                np.arange(0, ny, dtype=np.float64)
                * (ycrange[1] - ycrange[0])
                / (ny - 1)
                + ycrange[0]
            )
            regridded_zdata = interp_func(out_vdata_centers)

            # Now we'll convert from bin centers back to bin boundaries for pcolormesh
            # We're still in linear space at this point
            result = get_bin_boundaries(out_vdata_centers, ylog=False)
            rebinned_boundaries = result[0]

            if zlog:
                regridded_zdata = 10**regridded_zdata

            if ylog:
                bin_boundaries_1d = 10**rebinned_boundaries

    # check for negatives if zlog is requested
    if zlog:
        regridded_zdata[regridded_zdata < 0.0] = 0.0

    ylim_before = this_axis.get_ylim()
    # create the spectrogram (ignoring warnings)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        #logging.info("Starting pcolormesh")
        im = this_axis.pcolormesh(time_boundaries, bin_boundaries_1d.T, regridded_zdata.T, **spec_options)
        #logging.info("Done with pcolormesh")
    ylim_after = this_axis.get_ylim()

    #logging.info("ylim before pcolormesh: %s ",str(ylim_before))
    #logging.info("ylim after pcolormesh: %s", str(ylim_after))

    # store everything needed to create the colorbars
    colorbars[variable] = {}
    colorbars[variable]["im"] = im
    colorbars[variable]["axis_font_size"] = axis_font_size
    colorbars[variable]["ztitle"] = ztitle
    colorbars[variable]["zsubtitle"] = zsubtitle
    return True



from .load import load
from .lena import lena
from .mena import mena
from .hena import hena
from .rpi import rpi
from .euv import euv
from .fuv import fuv
from .orbit import orbit
from .datasets import datasets



import os

CONFIG = {"local_data_dir": "maven_data/", "maven_username": "", "maven_password": ""}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "maven"])

if os.environ.get("MAVEN_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ["MAVEN_DATA_DIR"]



"""Tests of ssl2dsl and dsl2gse functions."""

import unittest
from pyspedas.tplot_tools import get_data,del_data,tplot_restore,data_exists, set_coords
from numpy.testing import assert_array_almost_equal_nulp, assert_array_max_ulp, assert_allclose
from pyspedas.projects.themis import autoload_support, ssl2dsl,dsl2gse


class DSLCotransDataValidation(unittest.TestCase):
    """ Compares cotrans_tools results between Python and IDL """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: general/tools/python_validate/thm_cotrans_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.projects.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']

        # Testing tolerance
        cls.tol = 1e-10

        # Download tplot files
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        # remote_name = 'testfiles/thm_cotrans_validate.cdf'
        remote_name = 'cotrans_tools/thm_cotrans_validate.tplot'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path=CONFIG['local_data_dir'],
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        # cdf_to_tplot(filename)
        tplot_restore(filename)
        #tplot_names()
        cls.basis_x = get_data('basis_x')
        cls.basis_y = get_data('basis_y')
        cls.basis_z = get_data('basis_z')
        cls.basis_x_gei2gse = get_data('basis_x_gei2gse')
        cls.basis_y_gei2gse = get_data('basis_y_gei2gse')
        cls.basis_z_gei2gse = get_data('basis_z_gei2gse')
        cls.basis_x_gse2gei = get_data('basis_x_gse2gei')
        cls.basis_y_gse2gei = get_data('basis_y_gse2gei')
        cls.basis_z_gse2gei = get_data('basis_z_gse2gei')

        cls.basis_x_dsl2gse = get_data('basis_x_dsl2gse')
        cls.basis_y_dsl2gse = get_data('basis_y_dsl2gse')
        cls.basis_z_dsl2gse = get_data('basis_z_dsl2gse')

        cls.basis_x_gse2dsl = get_data('basis_x_gse2dsl')
        cls.basis_y_gse2dsl = get_data('basis_y_gse2dsl')
        cls.basis_z_gse2dsl = get_data('basis_z_gse2dsl')

        cls.basis_x_ssl2dsl = get_data('basis_x_ssl2dsl')
        cls.basis_y_ssl2dsl = get_data('basis_y_ssl2dsl')
        cls.basis_z_ssl2dsl = get_data('basis_z_ssl2dsl')

        cls.basis_x_dsl2ssl = get_data('basis_x_dsl2ssl')
        cls.basis_y_dsl2ssl = get_data('basis_y_dsl2ssl')
        cls.basis_z_dsl2ssl = get_data('basis_z_dsl2ssl')

        # The cotrans_tools routines now can load their own support data.  However, it seems you actually need
        # some substantial padding of the support data time interval compared to the target variable.  I
        # think this is due to the V03 state spinaxis and spinphase corrections, which are only given once per
        # day. So you need at least the preceding and following days to be able to interpolate to the current
        # date, or even two days to account for non-linear interpolation methods.  Perhaps autoload_support
        # should be taking this into account and add some extra padding.  The basis test vectors have timestamps
        # during 2007-03-23, but the IDL test data was generated with state loaded from 2007-03-29 through
        # 2007-03-30.

        # original time range from IDL test data generation
        # state_trange = ['2007-03-20', '2007-03-30']

        # smallest time interval that gives acceptably identical results to IDL
        state_trange = ['2007-03-21','2007-03-25']
        autoload_support(trange=state_trange, probe='a', spinaxis=True, spinmodel=True)


    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # del_data('*')

    def test_gei2gse(self):
        """Validate gei2gse transform """
        from pyspedas.cotrans_tools.cotrans_lib import subgei2gse

        bx = self.basis_x
        by = self.basis_y
        bz = self.basis_z
        times = bx.times
        bx_gse = subgei2gse(times, bx.y)
        by_gse = subgei2gse(times, by.y)
        bz_gse = subgei2gse(times, bz.y)
        assert_allclose(bx_gse, self.basis_x_gei2gse.y, atol=1.0e-06)
        assert_allclose(by_gse, self.basis_y_gei2gse.y, atol=1.0e-06)
        assert_allclose(bz_gse, self.basis_z_gei2gse.y, atol=1.0e-06)

    def test_gse2gei(self):
        """Validate gse2gei transform """
        from pyspedas.cotrans_tools.cotrans_lib import subgse2gei

        bx = self.basis_x
        by = self.basis_y
        bz = self.basis_z
        times = bx.times
        bx_gei = subgse2gei(times, bx.y)
        by_gei = subgse2gei(times, by.y)
        bz_gei = subgse2gei(times, bz.y)
        assert_allclose(bx_gei, self.basis_x_gse2gei.y, atol=1.0e-06)
        assert_allclose(by_gei, self.basis_y_gse2gei.y, atol=1.0e-06)
        assert_allclose(bz_gei, self.basis_z_gse2gei.y, atol=1.0e-06)

    def test_dsl2gse_x(self):
        """Validate dsl2gse X axis transform """

        set_coords('basis_x', 'DSL')
        result = dsl2gse('basis_x', 'basis_x_dsl2gse', probe='a')
        self.assertEqual(result,1)
        bx_gse = get_data('basis_x_dsl2gse')
        assert_allclose(bx_gse.y, self.basis_x_dsl2gse.y, atol=1.0e-06)

    def test_dsl2gse_y(self):
        """Validate dsl2gse Y axis transform """
        from pyspedas.cotrans_tools.cotrans_lib import subgse2gei
        set_coords('basis_y', 'DSL')
        result = dsl2gse('basis_y', 'basis_y_dsl2gse', probe='a')
        self.assertEqual(result, 1)
        by_gse = get_data('basis_y_dsl2gse')
        assert_allclose(by_gse.y, self.basis_y_dsl2gse.y, atol=1.0e-06)

    def test_dsl2gse_z(self):
        """Validate dsl2gse Z axis transform """
        set_coords('basis_z', 'DSL')
        result = dsl2gse('basis_z','basis_z_dsl2gse', probe='a')
        self.assertEqual(result, 1)
        bz_gse = get_data('basis_z_dsl2gse')
        assert_allclose(bz_gse.y, self.basis_z_dsl2gse.y, atol=1.0e-06)

    def test_gse2dsl_x(self):
        """Validate gse2dsl X axis transform """
        set_coords('basis_x', 'GSE')
        result = dsl2gse('basis_x', 'basis_x_gse2dsl', probe='a', isgsetodsl=True)
        self.assertEqual(result, 1)
        self.assertEqual(result, 1)
        bx_gse = get_data('basis_x_gse2dsl')
        assert_allclose(bx_gse.y, self.basis_x_gse2dsl.y, atol=1.0e-06)

    def test_gse2dsl_y(self):
        """Validate gse2dsl Y axis transform """
        from pyspedas.cotrans_tools.cotrans_lib import subgse2gei
        set_coords('basis_y', 'GSE')
        result = dsl2gse('basis_y', 'basis_y_gse2dsl', probe='a', isgsetodsl=True)
        self.assertEqual(result, 1)
        by_gse = get_data('basis_y_gse2dsl')
        assert_allclose(by_gse.y, self.basis_y_gse2dsl.y, atol=1.0e-06)

    def test_gse2dsl_z(self):
        """Validate gse2dsl Z axis transform """
        set_coords('basis_z', 'GSE')
        result = dsl2gse('basis_z', 'basis_z_gse2dsl', probe='a', isgsetodsl=True)
        self.assertEqual(result, 1)
        bz_gse = get_data('basis_z_gse2dsl')
        assert_allclose(bz_gse.y, self.basis_z_gse2dsl.y, atol=1.0e-06)

    def test_ssl2dsl_x(self):
        """Validate ssl2dsl X axis transform """
        set_coords('basis_x', 'SSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_x','basis_x_ssldsl',probe='a',eclipse_correction_level=1)
        self.assertEqual(result, 1)
        bx_dsl = get_data('basis_x_ssl2dsl')
        assert_allclose(bx_dsl.y, self.basis_x_ssl2dsl.y, atol=1.0e-06)

    def test_ssl2dsl_y(self):
        """Validate ssl2dsl Y axis transform """
        set_coords('basis_y', 'SSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_y','basis_y_ssldsl',probe='a',eclipse_correction_level=1, use_spinphase_correction=True)
        self.assertEqual(result, 1)
        by_dsl = get_data('basis_y_ssl2dsl')
        assert_allclose(by_dsl.y, self.basis_y_ssl2dsl.y, atol=1.0e-06)

    def test_ssl2dsl_z(self):
        """Validate ssl2dsl Z axis transform """
        set_coords('basis_z', 'SSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_z','basis_z_ssldsl', probe='a',eclipse_correction_level=1,use_spinphase_correction=True)
        self.assertEqual(result, 1)
        bz_dsl = get_data('basis_z_ssl2dsl')
        assert_allclose(bz_dsl.y, self.basis_z_ssl2dsl.y, atol=1.0e-06)

    def test_dsl2ssl_x(self):
        """Validate dsl2ssl X axis transform """
        set_coords('basis_x', 'DSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_x','basis_x_dsl2ssl',probe='a',eclipse_correction_level=1, use_spinphase_correction=True, isdsltossl=True)
        self.assertEqual(result, 1)
        bx_ssl = get_data('basis_x_dsl2ssl')
        # This test needs a slightly looser tolerance for some reason.
        assert_allclose(bx_ssl.y, self.basis_x_dsl2ssl.y, atol=1.5e-06)

    def test_dsl2ssl_y(self):
        """Validate dsl2ssl Y axis transform """
        set_coords('basis_y', 'DSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_y','basis_y_dsl2ssl', probe='a', eclipse_correction_level=1, use_spinphase_correction=True, isdsltossl=True)
        self.assertEqual(result, 1)
        by_ssl = get_data('basis_y_dsl2ssl')
        # This test needs a slightly looser tolerance for some reason.
        assert_allclose(by_ssl.y, self.basis_y_dsl2ssl.y, atol=1.5e-06)

    def test_dsl2ssl_z(self):
        """Validate dsl2ssl Z axis transform """
        set_coords('basis_z', 'DSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_z','basis_z_dsl2ssl', probe='a', eclipse_correction_level=1, use_spinphase_correction=True, isdsltossl=True)
        self.assertEqual(result, 1)
        bz_ssl = get_data('basis_z_dsl2ssl')
        assert_allclose(bz_ssl.y, self.basis_z_dsl2ssl.y, atol=1.0e-06)

    def test_catch_mismatch_dsl2ssl_z(self):
        """Test detection of mismatched input vs. requested coordinate systems in dsl2ssl transform """
        # Requesting DSL to SSL, but specifying SSL as input coordinate system
        set_coords('basis_z', 'SSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_z','basis_z_dsl2ssl', probe='a', eclipse_correction_level=1, use_spinphase_correction=True, isdsltossl=True, ignore_input_coord = False)
        self.assertEqual(result, 0)

    def test_catch_mismatch_ssl2dsl_z(self):
        """Test detection of mismatched input vs. requested coordinates in ssl2dsl transform """
        # Requesting SSL to DSL, but specifying DSL as input coordinate system
        set_coords('basis_z', 'DSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_z','basis_z_ssldsl', probe='a', eclipse_correction_level=1, use_spinphase_correction=True, ignore_input_coord=False)
        self.assertEqual(result, 0)

    def test_catch_mismatch_gse2dsl_z(self):
        """Test detection of mismatched input vs requested coordinates in gse2dsl transform """
        set_coords('basis_z', 'DSL')
        result = dsl2gse('basis_z', 'basis_z_gse2dsl', probe='a', isgsetodsl=True)
        self.assertEqual(result, 0)

    def test_catch_mismatch_dsl2gse_z(self):
        """Test detection of mismatched input vs. requested coordinates in dsl2gse transform """
        set_coords('basis_z', 'GSE')
        result = dsl2gse('basis_z', 'basis_z_gse2dsl', probe='a')
        self.assertEqual(result, 0)

    def test_ignore_mismatch_dsl2ssl_z(self):
        """Test ability to bypass coordinate system consistency check in dsl2ssl transform """
        # Requesting DSL to SSL, but specifying SSL as input coordinate system
        set_coords('basis_z', 'SSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_z','basis_z_dsl2ssl', probe='a', eclipse_correction_level=1, use_spinphase_correction=True, isdsltossl=True, ignore_input_coord = True)
        self.assertEqual(result, 1)

    def test_ignore_mismatch_ssl2dsl_z(self):
        """Test ability to bypass coordinate system consistency check in ssl2dsl transform  """
        # Requesting SSL to DSL, but specifying DSL as input coordinate system
        set_coords('basis_z', 'DSL')
        # Usually probe can be inferred from the input variable name, but we need it here.
        result = ssl2dsl('basis_z','basis_z_ssldsl', probe='a', eclipse_correction_level=1,use_spinphase_correction=True, ignore_input_coord=True)
        self.assertEqual(result, 1)

    def test_ignore_mismatch_gse2dsl_z(self):
        """Test ability to bypass coordinate system consistency check in gse2dsl transform """
        set_coords('basis_z', 'DSL')
        result = dsl2gse('basis_z', 'basis_z_gse2dsl', probe='a', isgsetodsl=True, ignore_input_coord=True)
        self.assertEqual(result, 1)

    def test_ignore_mismatch_dsl2gse_z(self):
        """Test ability to bypass coordinate system consistency check in dsl2gse transform """
        set_coords('basis_z', 'GSE')
        result = dsl2gse('basis_z', 'basis_z_gse2dsl', probe='a', ignore_input_coord=True)
        self.assertEqual(result, 1)

if __name__ == '__main__':
    unittest.main()



from .load import load

# Refer to __init__.py for revision history before the function definitions were split out

def reptile(trange=['2013-11-5', '2013-11-6'],
        datatype='flux', 
        level='l2',
        prefix='',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    This function loads data from the Relativistic Electron and Proton Telescope integrated little experiment (REPTile)

    Parameters
    ----------

        trange : list of str
            time range of interest [starttime, endtime] with the format
            [YYYY-MM-DD, YYYY-MM-DD], or to specify more or less than a day,
            [YYYY-MM-DD/hh:mm:ss, YYYY-MM-DD/hh:mm:ss].
            Default: ['2013-11-5', '2013-11-6]

        datatype: str
            Data type; Valid options::

                'counts' for L1 data
                'flux' for L2 data

        level: str
            Data level; options: 'l1', 'l2' (default: l2)

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
    list of str
        List of tplot variables created.

    """
    return load(instrument='reptile', trange=trange, level=level, datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download)





import os
import unittest
import numpy as np
import logging
import pyspedas
from pyspedas.tplot_tools import data_exists, del_data
from pyspedas.projects.goes.config import CONFIG

global_display = False
outputdir = os.path.join(CONFIG["local_data_dir"], "idltestfiles")
if os.path.exists(outputdir) is False:
    os.makedirs(outputdir)


class LoadTestCases(unittest.TestCase):

    def test_downloadonly(self):
        del_data()
        mag_files = pyspedas.projects.goes.fgm(datatype="1min", downloadonly=True)
        self.assertTrue(os.path.exists(mag_files[0]))

    def test_load_orbit_data(self):
        del_data()
        orbit_vars = pyspedas.projects.goes.orbit(downloadonly=True)
        filename = os.path.basename(orbit_vars[0])
        self.assertEqual(filename, "goes15_ephemeris_ssc_20130101_v01.cdf")
        orbit_vars = pyspedas.projects.goes.orbit(notplot=True, time_clip=False)
        # This should return a dict of tplot structures, but we're getting a list of names, which don't exist as tplot variables.
        self.assertTrue(isinstance(orbit_vars, dict))
        self.assertTrue("g15_orbit_XYZ_GSM" in orbit_vars.keys())
        orbit_vars = pyspedas.projects.goes.orbit(time_clip=True)
        self.assertTrue("g15_orbit_XYZ_GSM" in orbit_vars)
        self.assertTrue("g15_orbit_XYZ_GSE" in orbit_vars)
        self.assertTrue("g15_orbit_XYZ_SM" in orbit_vars)
        self.assertTrue("g15_orbit_XYZ_GEO" in orbit_vars)

        self.assertTrue(data_exists("g15_orbit_XYZ_GSM"))
        self.assertTrue(data_exists("g15_orbit_XYZ_GSE"))
        self.assertTrue(data_exists("g15_orbit_XYZ_SM"))
        self.assertTrue(data_exists("g15_orbit_XYZ_GEO"))
        del_data("*")
        # Both notplot and time_clip specified: should return dictionary, no crash, no tplot variables created
        orbit_vars = pyspedas.projects.goes.orbit(time_clip=True, notplot=True)
        self.assertTrue(isinstance(orbit_vars, dict))
        self.assertFalse(data_exists("g15_orbit_XYZ_GSM"))
        self.assertTrue("g15_orbit_XYZ_GSM" in orbit_vars.keys())

    def test_load_1min_mag_data(self):
        del_data()
        mag_vars = pyspedas.projects.goes.fgm(datatype="1min")
        self.assertTrue("g15_fgm_BX_1" in mag_vars)
        self.assertTrue("g15_fgm_BY_1" in mag_vars)
        self.assertTrue("g15_fgm_BZ_1" in mag_vars)
        self.assertTrue(data_exists("g15_fgm_BX_1"))
        self.assertTrue(data_exists("g15_fgm_BY_1"))
        self.assertTrue(data_exists("g15_fgm_BZ_1"))

    def test_load_5min_mag_data(self):
        del_data()
        mag_vars = pyspedas.projects.goes.fgm(
            datatype="5min",
            probe="10",
            trange=["2000-07-01", "2000-07-02"],
            time_clip=True,
        )
        self.assertTrue("g10_fgm_ht" in mag_vars)
        self.assertTrue(data_exists("g10_fgm_ht"))

    def test_load_full_mag_data(self):
        del_data()
        mag_vars = pyspedas.projects.goes.fgm(datatype="512ms", suffix="_512")
        self.assertTrue("g15_fgm_BX_1_512" in mag_vars)
        self.assertTrue("g15_fgm_BY_1_512" in mag_vars)
        self.assertTrue("g15_fgm_BZ_1_512" in mag_vars)
        self.assertTrue(data_exists("g15_fgm_BX_1_512"))
        self.assertTrue(data_exists("g15_fgm_BY_1_512"))
        self.assertTrue(data_exists("g15_fgm_BZ_1_512"))

    def test_load_1min_epead_data(self):
        del_data()
        epead_vars = pyspedas.projects.goes.epead()
        self.assertTrue("g15_epead_E1E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E1E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_E1W_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E1W_UNCOR_FLUX"))
        self.assertTrue("g15_epead_E2E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E2E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P1E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P1E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P1W_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P1W_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P2E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P2E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_A1E_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A1E_FLUX"))
        self.assertTrue("g15_epead_A1W_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A1W_FLUX"))
        self.assertTrue("g15_epead_A2E_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A2E_FLUX"))

    def test_load_full_epead_data(self):
        del_data()
        epead_vars = pyspedas.projects.goes.epead(datatype="1min")
        self.assertTrue("g15_epead_E1E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E1E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_E1W_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E1W_UNCOR_FLUX"))
        self.assertTrue("g15_epead_E2E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E2E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P1E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P1E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P1W_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P1W_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P2E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P2E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_A1E_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A1E_FLUX"))
        self.assertTrue("g15_epead_A1W_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A1W_FLUX"))
        self.assertTrue("g15_epead_A2E_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A2E_FLUX"))

    def test_load_5min_epead_data(self):
        del_data()
        epead_vars = pyspedas.projects.goes.epead(datatype="5min")
        self.assertTrue("g15_epead_E1E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E1E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_E1W_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E1W_UNCOR_FLUX"))
        self.assertTrue("g15_epead_E2E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_E2E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P1E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P1E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P1W_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P1W_UNCOR_FLUX"))
        self.assertTrue("g15_epead_P2E_UNCOR_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_P2E_UNCOR_FLUX"))
        self.assertTrue("g15_epead_A1E_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A1E_FLUX"))
        self.assertTrue("g15_epead_A1W_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A1W_FLUX"))
        self.assertTrue("g15_epead_A2E_FLUX" in epead_vars)
        self.assertTrue(data_exists("g15_epead_A2E_FLUX"))

    def test_load_full_maged_data(self):
        del_data()
        maged_vars = pyspedas.projects.goes.maged(datatype="full")
        self.assertTrue("g15_maged_M_1ME1_DTC_UNCOR_CR" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME1_DTC_UNCOR_CR"))
        self.assertTrue("g15_maged_M_1ME2_DTC_UNCOR_CR" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME2_DTC_UNCOR_CR"))
        self.assertTrue("g15_maged_M_1ME3_DTC_UNCOR_CR" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME3_DTC_UNCOR_CR"))
        self.assertTrue("g15_maged_M_1ME4_DTC_UNCOR_CR" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME4_DTC_UNCOR_CR"))
        self.assertTrue("g15_maged_M_1ME5_DTC_UNCOR_CR" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME5_DTC_UNCOR_CR"))

    def test_load_1min_maged_data(self):
        del_data()
        maged_vars = pyspedas.projects.goes.maged(datatype="1min", time_clip=True)
        self.assertTrue("g15_maged_M_1ME1_DTC_UNCOR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME1_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_maged_M_1ME2_DTC_UNCOR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME2_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_maged_M_1ME3_DTC_UNCOR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME3_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_maged_M_1ME4_DTC_UNCOR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME4_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_maged_M_1ME5_DTC_UNCOR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_1ME5_DTC_UNCOR_FLUX"))

    def test_load_5min_maged_data(self):
        del_data()
        maged_vars = pyspedas.projects.goes.maged(datatype="5min", time_clip=True)
        self.assertTrue("g15_maged_M_2ME1_DTC_COR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_2ME1_DTC_COR_FLUX"))
        self.assertTrue("g15_maged_M_2ME2_DTC_COR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_2ME2_DTC_COR_FLUX"))
        self.assertTrue("g15_maged_M_2ME3_DTC_COR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_2ME3_DTC_COR_FLUX"))
        self.assertTrue("g15_maged_M_2ME4_DTC_COR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_2ME4_DTC_COR_FLUX"))
        self.assertTrue("g15_maged_M_2ME5_DTC_COR_FLUX" in maged_vars)
        self.assertTrue(data_exists("g15_maged_M_2ME5_DTC_COR_FLUX"))

    def test_load_full_magpd_data(self):
        del_data()
        magpd_vars = pyspedas.projects.goes.magpd(datatype="full")
        self.assertTrue("g15_magpd_M_1MP1_DTC_UNCOR_CR" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP1_DTC_UNCOR_CR"))
        self.assertTrue("g15_magpd_M_1MP2_DTC_UNCOR_CR" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP2_DTC_UNCOR_CR"))
        self.assertTrue("g15_magpd_M_1MP3_DTC_UNCOR_CR" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP3_DTC_UNCOR_CR"))
        self.assertTrue("g15_magpd_M_1MP4_DTC_UNCOR_CR" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP4_DTC_UNCOR_CR"))
        self.assertTrue("g15_magpd_M_1MP5_DTC_UNCOR_CR" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP5_DTC_UNCOR_CR"))

    def test_load_1min_magpd_data(self):
        del_data()
        magpd_vars = pyspedas.projects.goes.magpd(datatype="1min", time_clip=True)
        self.assertTrue("g15_magpd_M_1MP1_DTC_UNCOR_FLUX" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP1_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_magpd_M_1MP2_DTC_UNCOR_FLUX" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP2_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_magpd_M_1MP3_DTC_UNCOR_FLUX" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP3_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_magpd_M_1MP4_DTC_UNCOR_FLUX" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP4_DTC_UNCOR_FLUX"))
        self.assertTrue("g15_magpd_M_1MP5_DTC_UNCOR_FLUX" in magpd_vars)
        self.assertTrue(data_exists("g15_magpd_M_1MP5_DTC_UNCOR_FLUX"))

    def test_load_full_hepad_data(self):
        del_data()
        hepad_vars = pyspedas.projects.goes.hepad(datatype="full")
        self.assertTrue("g15_hepad_P10_FLUX" in hepad_vars)
        self.assertTrue("g15_hepad_S1_COUNT_RATE" in hepad_vars)
        self.assertTrue(data_exists("g15_hepad_P10_FLUX"))
        self.assertTrue(data_exists("g15_hepad_S1_COUNT_RATE"))

    def test_load_1min_hepad_data(self):
        del_data()
        hepad_vars = pyspedas.projects.goes.hepad(prefix="probename", time_clip=True)
        self.assertTrue("g15_hepad_P10_FLUX" in hepad_vars)
        self.assertTrue("g15_hepad_S1_COUNT_RATE" in hepad_vars)
        self.assertTrue(data_exists("g15_hepad_P10_FLUX"))
        self.assertTrue(data_exists("g15_hepad_S1_COUNT_RATE"))

    def test_load_xrs_data(self):
        del_data()
        xrs_vars = pyspedas.projects.goes.xrs(
            probe="10", datatype="full", trange=["2002-08-01", "2002-08-01"]
        )
        self.assertTrue("g10_xrs_xl" in xrs_vars)
        self.assertTrue(data_exists("g10_xrs_xl"))

    def test_load_xrs_5m_data(self):
        del_data()
        xrs_vars = pyspedas.projects.goes.xrs(
            probe="11",
            datatype="5min",
            trange=["2000-09-01", "2000-09-01"],
            time_clip=True,
        )
        self.assertTrue("g11_xrs_xl" in xrs_vars)
        self.assertTrue(data_exists("g11_xrs_xl"))

    def test_load_xrs_1m_data(self):
        del_data()
        xrs_vars = pyspedas.projects.goes.xrs(
            probe="11",
            datatype="1min",
            trange=["2000-09-01", "2000-09-01"],
            prefix="probename",
            time_clip=True,
        )
        self.assertTrue("g11_xrs_xs" in xrs_vars)
        self.assertTrue(data_exists("g11_xrs_xs"))

    def test_fillval_removal(self):
        del_data()
        trange = ["2024-05-21", "2024-05-22"]
        probe = 18

        pyspedas.projects.goes.euvs(probe=probe, trange=trange)
        pyspedas.projects.goes.xrs(probe=probe, trange=trange)
        local_png = os.path.join(outputdir, "testing_g18_xrs_euvs.png")
        pyspedas.tplot(
            ["g18_euvs_MgII_standard", "g18_xrs_xrsa_flux"],
            display=global_display,
            save_png=local_png,
        )
        d = pyspedas.get_data("g18_euvs_MgII_standard")
        datamin = np.nanmin(d.y)
        # Fillval should be -9999 for this variable, was it removed?
        self.assertTrue(datamin > -1.0)

    def test_xrs_scale(self):
        del_data()
        trange = ["2024-05-21", "2024-05-22"]
        probe = 18

        pyspedas.projects.goes.xrs(probe=probe, trange=trange)
        local_png = os.path.join(outputdir, "testing_g18_xrs_xrsa_flux.png")
        pyspedas.tplot(
            ["g18_xrs_xrsa_flux"],
            display=global_display,
            save_png=local_png,
        )
        md = pyspedas.get_data("g18_xrs_xrsa_flux", metadata=True)
        scale = md["plot_options"]["yaxis_opt"]["y_axis_type"]
        self.assertTrue(scale == "log")

    def test_load_eps_1m_data(self):
        del_data()
        eps_vars = pyspedas.projects.goes.eps(
            trange=["2000-09-01", "2000-09-01"], probe="11", time_clip=True
        )
        self.assertTrue("g11_eps_e1_flux_i" in eps_vars)
        self.assertTrue("g11_eps_e2_flux_i" in eps_vars)
        self.assertTrue("g11_eps_e3_flux_i" in eps_vars)
        self.assertTrue(data_exists("g11_eps_e1_flux_i"))
        self.assertTrue(data_exists("g11_eps_e2_flux_i"))
        self.assertTrue(data_exists("g11_eps_e3_flux_i"))

    def test_load_eps_5m_data(self):
        del_data()
        eps_vars = pyspedas.projects.goes.eps(
            trange=["2000-09-01", "2000-09-01"],
            probe="11",
            datatype="5min",
            time_clip=True,
        )
        self.assertTrue("g11_eps_p1_flux" in eps_vars)
        self.assertTrue("g11_eps_p2_flux" in eps_vars)
        self.assertTrue("g11_eps_p3_flux" in eps_vars)
        self.assertTrue("g11_eps_p4_flux" in eps_vars)
        self.assertTrue("g11_eps_p5_flux" in eps_vars)
        self.assertTrue("g11_eps_p6_flux" in eps_vars)
        self.assertTrue("g11_eps_p7_flux" in eps_vars)
        self.assertTrue(data_exists("g11_eps_p1_flux"))
        self.assertTrue(data_exists("g11_eps_p2_flux"))
        self.assertTrue(data_exists("g11_eps_p3_flux"))
        self.assertTrue(data_exists("g11_eps_p4_flux"))
        self.assertTrue(data_exists("g11_eps_p5_flux"))
        self.assertTrue(data_exists("g11_eps_p6_flux"))
        self.assertTrue(data_exists("g11_eps_p7_flux"))

    def test_load_xrs_data_16(self):
        del_data()
        xrs_vars_16 = pyspedas.projects.goes.xrs(
            probe="16", trange=["2022-09-01", "2022-09-02"]
        )
        self.assertTrue("g16_xrs_xrsa_flux" in xrs_vars_16)
        self.assertTrue(data_exists("g16_xrs_xrsa_flux"))

    def test_load_xrs_data_17_hi(self):
        del_data()
        xrs_vars_17 = pyspedas.projects.goes.xrs(
            probe="17", trange=["2022-07-01", "2022-07-02"], datatype="hi"
        )
        self.assertTrue("g17_xrs_xrsb_flux" in xrs_vars_17)
        self.assertTrue(data_exists("g17_xrs_xrsb_flux"))

    def test_load_euvs_data_17(self):
        del_data()
        euvs_vars_17 = pyspedas.projects.goes.euvs(
            probe="17", trange=["2022-09-01", "2022-09-02"]
        )
        self.assertTrue("g17_euvs_irr_256" in euvs_vars_17)
        self.assertTrue(data_exists("g17_euvs_irr_256"))

    def test_load_euvs_data_16_hi(self):
        del_data()
        euvs_vars_16 = pyspedas.projects.goes.euvs(
            probe="16",
            trange=["2022-08-01", "2022-08-02"],
            prefix="probename",
            datatype="hi",
        )
        self.assertTrue("g16_euvs_irr_256" in euvs_vars_16)
        self.assertTrue(data_exists("g16_euvs_irr_256"))

    def test_load_mag_data_16(self):
        del_data()
        mag_vars_16 = pyspedas.projects.goes.mag(
            probe="16", trange=["2023-01-30", "2023-01-31"], prefix="goes16_"
        )
        self.assertTrue("goes16_b_total" in mag_vars_16)
        self.assertTrue(data_exists("goes16_b_total"))

    def test_load_mag_data_17(self):
        del_data()
        mag_vars_17 = pyspedas.projects.goes.mag(
            probe="17",
            trange=["2022-01-30", "2022-01-31"],
            prefix="goes17_",
            datatype="hi",
        )
        self.assertTrue("goes17_b_gse" in mag_vars_17)
        self.assertTrue(data_exists("goes17_b_gse"))

    def test_load_mpsh_data_17(self):
        del_data()
        mpsh_vars_17 = pyspedas.projects.goes.mpsh(
            probe="17",
            trange=["2022-09-01", "2022-09-02"],
            prefix="probename",
            time_clip=True,
        )
        self.assertTrue("g17_mpsh_AvgDiffElectronFlux" in mpsh_vars_17)
        self.assertTrue(data_exists("g17_mpsh_AvgDiffElectronFlux"))

    def test_load_mpsh_data_16(self):
        del_data()
        mpsh_vars_16 = pyspedas.projects.goes.mpsh(
            probe="16", trange=["2022-10-01", "2022-10-02"], prefix="", datatype="hi"
        )
        self.assertTrue("g16_mpsh_AvgDiffProtonFlux" in mpsh_vars_16)
        self.assertTrue(data_exists("g16_mpsh_AvgDiffProtonFlux"))

    def test_load_sgps_data_16(self):
        del_data()
        sgps_vars_16 = pyspedas.projects.goes.sgps(
            probe="16", trange=["2023-01-30", "2023-01-31"], prefix="probename"
        )
        self.assertTrue("g16_sgps_AvgDiffAlphaFlux" in sgps_vars_16)
        self.assertTrue(data_exists("g16_sgps_AvgDiffAlphaFlux"))

    def test_load_sgps_data_18(self):
        del_data()
        sgps_vars_18 = pyspedas.projects.goes.sgps(
            probe="18",
            trange=["2023-01-30", "2023-01-31"],
            datatype="hi",
            time_clip=True,
        )
        self.assertTrue("g18_sgps_AvgIntProtonFlux" in sgps_vars_18)
        self.assertTrue(data_exists("g18_sgps_AvgIntProtonFlux"))

    def test_load_goes_reprocessed(self):
        # Load reprocessed data, GOES 8-15, GOES-R file format

        # 1. hires mag (large file, slow)
        del_data()
        mag_vars = pyspedas.projects.goes.load(
            probe="13",
            instrument="mag",
            trange=["2015-01-01 00:00:00", "2015-01-01 12:00:00"],
            time_clip=True,
            goes_r=True,
        )
        logging.info(mag_vars)
        self.assertTrue("g13_mag_b_gsm" in mag_vars)
        # tplot("g13_mag_b_gsm")

        # 2. hires xrs
        del_data()
        xrs_vars = pyspedas.projects.goes.load(
            probe="13",
            instrument="xrs",
            datatype="hi",
            trange=["2015-01-20 00:00:00", "2015-01-20 12:00:00"],
            time_clip=True,
            goes_r=True,
        )
        logging.info(xrs_vars)
        self.assertTrue("g13_xrs_a_flux" in xrs_vars)
        self.assertTrue("g13_xrs_b_flux" in xrs_vars)
        # tplot(["g13_xrs_a_flux", "g13_xrs_b_flux"])

        # 3. lowres xrs
        # del_data()
        xrs_vars2 = pyspedas.projects.goes.load(
            probe="13",
            instrument="xrs",
            trange=["2015-01-20 00:00:00", "2015-01-20 12:00:00"],
            time_clip=True,
            goes_r=True,
        )
        logging.info(xrs_vars2)
        self.assertTrue("g13_xrs_xrsa_flux" in xrs_vars2)
        self.assertTrue("g13_xrs_xrsb_flux" in xrs_vars2)
        # tplot(["g13_xrs_xrsa_flux", "g13_xrs_xrsb_flux"])

        # The following shows both hi and lowres xrs data together
        # tplot(["g13_xrs_a_flux", "g13_xrs_xrsa_flux","g13_xrs_b_flux","g13_xrs_xrsb_flux"])


if __name__ == "__main__":
    unittest.main()



import cdflib
import numpy as np

from copy import deepcopy
from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load
from ...satellite.erg.get_gatt_ror import get_gatt_ror


from typing import List, Optional, Union


def isee_brio(
    trange: List[str] = ["2020-08-01", "2020-08-02"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    Load isee_brio riometer data from ERG Science Center

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2020-08-01', '2020-08-02']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load. Valid values: 'ath', 'kap', 'gak', 'hus', 'zgn', 'ist', 'all'
            Default: ['all']

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False


    Returns
    -------

    Examples
    ________
    >>> import pyspedas
    >>> brio_vars=pyspedas.projects.erg.isee_brio(trange=['2020-08-01', '2020-08-02'],site='ath')
    >>> print(brio_vars)

    """

    # ;----- datatype -----;
    datatype = "64hz"
    instr = "brio"
    freq = "30"

    site_code_all = ["ath", "kap", "gak", "hus", "zgn", "ist"]
    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(" ")
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if "all" in site_code:
        site_code = site_code_all

    site_code = list(set(site_code).intersection(site_code_all))

    new_cdflib = False
    if cdflib.__version__ > "0.4.9":
        new_cdflib = True
    else:
        new_cdflib = False

    prefix = "iseetmp_"
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        fres = datatype
        file_res = 3600.0 * 24
        pathformat = (
            "ground/riometer/"
            + site_input
            + "/%Y/isee_"
            + fres
            + "_"
            + instr
            + freq
            + "_"
            + site_input
            + "_%Y%m%d_v??.cdf"
        )

        loaded_data_temp = load(
            pathformat=pathformat,
            file_res=file_res,
            trange=trange,
            datatype=datatype,
            prefix=prefix,
            suffix="_" + site_input + suffix,
            get_support_data=get_support_data,
            varformat=varformat,
            downloadonly=downloadonly,
            notplot=notplot,
            time_clip=time_clip,
            no_update=no_update,
            uname=uname,
            passwd=passwd,
            force_download=force_download,
        )

        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                gatt = get_gatt_ror(downloadonly, loaded_data)
                print(
                    "**************************************************************************"
                )
                print(gatt["Logical_source_description"])
                print("")
                print(f'Information about {gatt["Station_code"]}')
                print(f'PI {gatt["PI_name"]}')
                print("")
                print(f'Affiliations: {gatt["PI_affiliation"]}')
                print("")
                print("Rules of the Road for ISEE Riometer Data:")
                print("")
                print(gatt["TEXT"])
                print(gatt["LINK_TEXT"])
                print(
                    "**************************************************************************"
                )
            except:
                print("printing PI info and rules of the road was failed")

        if (not downloadonly) and (not notplot):
            if len(loaded_data_temp) > 0:
                file_name = get_data(loaded_data_temp[-1], metadata=True)["CDF"][
                    "FILENAME"
                ]
                if isinstance(file_name, list):
                    file_name = file_name[0]
                cdf_file = cdflib.CDF(file_name)
                cdf_info = cdf_file.cdf_info()
                if new_cdflib:
                    all_cdf_variables = cdf_info.rVariables + cdf_info.zVariables
                else:
                    all_cdf_variables = cdf_info["rVariables"] + cdf_info["zVariables"]
                for t_plot_name in loaded_data_temp:
                    get_data_vars = get_data(t_plot_name)
                    if get_data_vars is None:
                        store_data(t_plot_name, delete=True)
                    else:
                        t_plot_name_split = t_plot_name.split("_")
                        if len(t_plot_name_split) > 2:
                            # ;----- Find param -----;
                            param = t_plot_name_split[1]
                            if param in ["cna", "qdc", "raw"]:
                                # ;----- Rename tplot variables -----;
                                new_tplot_name = (
                                    "isee_brio"
                                    + freq
                                    + "_"
                                    + site_input
                                    + "_"
                                    + fres
                                    + "_"
                                    + param
                                    + suffix
                                )
                                store_data(t_plot_name, newname=new_tplot_name)
                                loaded_data.remove(t_plot_name)
                                loaded_data.append(new_tplot_name)
                                # ;----- Missing data -1.e+31 --> NaN -----;
                                clip(new_tplot_name, -1e5, 1e5)
                                get_data_vars = get_data(new_tplot_name)
                                if param in all_cdf_variables:
                                    var_atts = cdf_file.varattsget(param)
                                    if (
                                        "FILLVAL" in var_atts
                                    ):  # removing "FILLVAL", like 999.9000 or 0.0.
                                        # removing process of cdf_to_tplot.py may be not working well.
                                        var_properties = cdf_file.varinq(param)
                                        if new_cdflib:
                                            dtdesc = var_properties.Data_Type_Description
                                        else:
                                            dtdesc = var_properties["Data_Type_Description"]
                                        if (
                                            (dtdesc == "CDF_FLOAT")
                                            or (dtdesc == "CDF_REAL4")
                                            or (dtdesc == "CDF_DOUBLE")
                                            or (dtdesc == "CDF_REAL8")
                                        ):
                                            if isinstance(var_atts["FILLVAL"], str):
                                                fill_value = float(var_atts["FILLVAL"])
                                            else:
                                                fill_value = deepcopy(
                                                    var_atts["FILLVAL"]
                                                )
                                            new_y = np.where(
                                                get_data_vars[1] == fill_value,
                                                np.nan,
                                                get_data_vars[1],
                                            )
                                            get_metadata_vars = get_data(
                                                new_tplot_name, metadata=True
                                            )
                                            store_data(
                                                new_tplot_name,
                                                data={
                                                    "x": get_data_vars[0],
                                                    "y": new_y,
                                                },
                                                attr_dict=get_metadata_vars,
                                            )
                                            ylim(
                                                new_tplot_name,
                                                np.nanmin(new_y),
                                                np.nanmax(new_y),
                                            )
                                    else:
                                        ylim(
                                            new_tplot_name,
                                            np.nanmin(get_data_vars[1]),
                                            np.nanmax(get_data_vars[1]),
                                        )
                                # ;----- Set options -----;
                                options(new_tplot_name, "ytitle", site_input.upper())
                                if param == "cna":
                                    options(new_tplot_name, "ysubtitle", "[dB]")
                                    options(new_tplot_name, "legend_names", ["CNA"])
                                elif param == "qdc":
                                    options(new_tplot_name, "ysubtitle", "[V]")
                                    options(new_tplot_name, "legend_names", ["QDC"])
                                elif param == "raw":
                                    options(new_tplot_name, "ysubtitle", "[V]")
                                    options(
                                        new_tplot_name, "legend_names", ["Raw data"]
                                    )

    return loaded_data



from .load import load

# Refer to __init__.py for revision history before load routines split out into separate files

def cindi(trange=['2013-11-5', '2013-11-6'],
        prefix='',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    This function loads data from the Coupled Ion-Neutral Dynamics Investigation (CINDI)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2013-11-5', '2013-11-6']

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    ----------
        List of tplot variables created.

    Example:
    ----------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> cindi_vars = pyspedas.projects.cnofs.cindi(trange=['2013-11-5', '2013-11-6'])
        >>> tplot(['ionVelocityX', 'ionVelocityY', 'ionVelocityZ'])

    """
    return load(instrument='cindi', trange=trange, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download)





from .load import load
from functools import update_wrapper
from pyspedas.utilities.pyspedas_functools import better_partial
from pyspedas.utilities.datasets import find_datasets


# Define partial wrappers for other load routines, fixing the instrument parameter
# 'better_partial' works better with PyCharm autocompletion than functools.partial
# update_wrapper() is necessary for help() to show info for the wrapped function,
# rather than the partial() object.

mag = better_partial(load, instrument="mag")
update_wrapper(mag, load)
nacs = better_partial(load, instrument="nacs")
update_wrapper(nacs, load)
rpa = better_partial(load, instrument="rpa")
update_wrapper(rpa, load)
fpi = better_partial(load, instrument="fpi")
update_wrapper(fpi, load)
idm = better_partial(load, instrument="idm")
update_wrapper(idm, load)
wats = better_partial(load, instrument="wats")
update_wrapper(wats, load)
vefi = better_partial(load, instrument="vefi")
update_wrapper(vefi, load)
lang = better_partial(load, instrument="lang")
update_wrapper(lang, load)
datasets = better_partial(find_datasets, mission="Dynamics Explorer", label=True)
update_wrapper(datasets, find_datasets)



def get_var_label_ticks(var_xr, times):
    out_ticks = []
    extras = var_xr.attrs['plot_options']['extras']
    var_label_format = extras.get('var_label_format')
    if var_label_format is None:
        var_label_format = '{:.2f}'
    for time in times:
        out_ticks.append(var_label_format.format(var_xr.interp(coords={'time': time}, kwargs={'fill_value': 'extrapolate', 'bounds_error': False}).values))
    return out_ticks



"""
Time clip of data.

Notes
-----
Similar to tclip.pro in IDL SPEDAS.

"""
import logging
import pyspedas
from pyspedas.tplot_tools import store_data, get_data, tnames, time_float, time_string, tplot_copy
import numpy as np
import copy

def time_clip(
        names,
        time_start,
        time_end,
        newname=None,
        suffix='-tclip',
        overwrite=False,
        interior_clip=False
):
    """
    Clip data from time_start to time_end.

    Parameters
    ----------
    names: str/list of str
        List of tplot variable names to time clip (wildcards accepted)
    time_start : float or string
        Start time.
    time_end : float or string
        End time.
    newname: str/list of str, optional
        List of new names for tplot variables.
        Default: None. If not given, then a suffix is applied or the variables are overwritten
    suffix: str, optional
        A suffix to apply.
        Default: '-tclip'
    overwrite: bool, optional
        If true, overwrite the existing tplot variable.
        Default: False
    interior_clip: bool, optional
        If true, reverse sense of operation and clip out times within the start/end range, for example,
        when manually despiking data.
        Default: False

    Returns
    -------
    list of str
        Returns a list of tplot variables created or changed

    Example
    -------
        >>> # Clip time
        >>> import pyspedas
        >>> x1 = [0, 4, 8, 12, 16]
        >>> time1 = [pyspedas.time_float("2020-01-01") + i for i in x1]
        >>> pyspedas.store_data("a", data={"x": time1, "y": [[1, 2, 3],[2, 3, 4],[3, 4, 5],[4, 5, 6],[5, 6,7]]})
        >>> time_start=time1[0]
        >>> time_end=time1[2]
        >>> pyspedas.time_clip('a',time_start,time_end)
        >>> ac = pyspedas.get_data('a-tclip')
        >>> print(ac)

    """
    if len(names) < 1:
        logging.warning('time_clip: no valid tplot variables specified')
        return

    old_names = tnames(names)

    if len(old_names) < 1:
        logging.warning('time_clip: No valid tplot variables matching '+str(names))
        return

    if overwrite:
        n_names = old_names
    elif (newname is None) or (len(newname) < 1) or (newname == ''):
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        logging.warning('time_clip: new names and input names have different lengths, using suffixes instead')
        n_names = [s + suffix for s in old_names]

    # We want the start/end times as both strings (for log messages) and floats (for comparisons)

    if isinstance(time_start, str):
        time_start_str = time_start
        time_start_float = time_float(time_start)
    else:
        time_start_str = time_string(time_start)
        time_start_float = time_start

    if isinstance(time_end, str):
        time_end_str = time_end
        time_end_float = time_float(time_end)
    else:
        time_end_str = time_string(time_end)
        time_end_float = time_end


    if time_start_float > time_end_float:
        logging.error('time_clip: Start time ' + time_start_str +' is larger than end time ' + time_end_str)
        return

    for j in range(len(old_names)):
        if old_names[j] != n_names[j]:
            tplot_copy(old_names[j], n_names[j])

        alldata = get_data(n_names[j])
        metadata = copy.deepcopy(get_data(n_names[j], metadata=True))

        if not isinstance(alldata, tuple): # NRV variable
            continue
            
        time = alldata[0]
        data = alldata[1]
        input_count = len(time)


        if input_count < 1:
            logging.info('time_clip found empty data for variable '+old_names[j])
            continue

        new_time = np.array(time_float(time))

        if interior_clip:
            # Invert sense of default comparison
            cond = (new_time < time_start_float) | (new_time > time_end_float)
        else:
            cond = (new_time >= time_start_float) & (new_time <= time_end_float)

        count = np.count_nonzero(cond)

        if count == 0:
            logging.warning('time_clip: '+ old_names[j] + ' has no data in requested range')
            continue

        if count == input_count:
            logging.debug('Time clip returns full data set for variable '+old_names[j])
            continue

        tmp_q = pyspedas.tplot_tools.data_quants[n_names[j]]

        if 'v1' in tmp_q.coords.keys():
            if len(tmp_q.coords['v1'].values.shape) == 2:
                v1_data = tmp_q.coords['v1'].values[cond, :]
            else:
                v1_data = tmp_q.coords['v1'].values

        if 'v2' in tmp_q.coords.keys():
            if len(tmp_q.coords['v2'].values.shape) == 2:
                v2_data = tmp_q.coords['v2'].values[cond, :]
            else:
                v2_data = tmp_q.coords['v2'].values

        if 'v3' in tmp_q.coords.keys():
            if len(tmp_q.coords['v3'].values.shape) == 2:
                v3_data = tmp_q.coords['v3'].values[cond, :]
            else:
                v3_data = tmp_q.coords['v3'].values

        if 'v' in tmp_q.coords.keys():
            if len(tmp_q.coords['v'].values.shape) == 2:
                v_data = tmp_q.coords['v'].values[cond, :]
            else:
                v_data = tmp_q.coords['v'].values

        if 'spec_bins' in tmp_q.coords.keys():
            if len(tmp_q.coords['spec_bins'].values.shape) == 2:
                v_data = tmp_q.coords['spec_bins']\
                    .values[cond, :]
            else:
                v_data = tmp_q.coords['spec_bins'].values

        try:
            if 'v1' in tmp_q.coords.keys() and\
                'v2' in tmp_q.coords.keys() and\
                    'v3' in tmp_q.coords.keys():
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond, :, :, :],
                    'v1': v1_data, 'v2': v2_data, 'v3': v3_data},
                    attr_dict=metadata)
            elif 'v1' in tmp_q.coords.keys() and\
                    'v2' in tmp_q.coords.keys():
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond, :, :],
                    'v1': v1_data, 'v2': v2_data},
                    attr_dict=metadata)
            elif 'v1' in tmp_q.coords.keys():
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond, :],
                    'v1': v1_data}, attr_dict=metadata)
            elif 'spec_bins' in tmp_q.coords.keys():
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond, :],
                    'v': v_data}, attr_dict=metadata)
            elif 'v' in tmp_q.coords.keys():
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond, :],
                    'v': v_data}, attr_dict=metadata)
            elif data.ndim == 1:
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond]},
                    attr_dict=metadata)
            else:
                store_data(n_names[j], data={
                    'x': time[cond],
                    'y': data[cond]},
                    attr_dict=metadata)
        except Exception as e:
            logging.error('Problem time clipping: ' + n_names[j])
            logging.error('Exception:'+str(e))
            continue

        logging.debug('Time clip was applied to: ' + n_names[j])

    return n_names



import numpy as np
from pyspedas.tplot_tools import store_data, options
from pyspedas.tplot_tools import time_double
from pyspedas.cotrans_tools.xyz_to_polar import xyz_to_polar
from .load_csv_file import load_csv_file


# This routine was originally in akebono/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def orb_postprocessing(files, prefix='', suffix=''):
    """
    Load the orbit CSV files and create the tplot variables

    Parameters
    ----------
    files: list of str
        List of CSV files to convert to tplot variables
    prefix: str
         A prefix to be added to the tplot variables created. Default: ''
    suffix: str
        A suffix to be added to the tplot variables created. Default: ''

    Returns
    -------
    list of str
        List of tplot variables created.
    """

    if prefix is None:
        user_prefix = ''
    else:
        user_prefix = prefix

    if suffix is None:
        suffix = ''

    prefix_project = 'akb_'
    prefix_descriptor = 'orb_'
    prefix = user_prefix + prefix_project + prefix_descriptor

    cols = ['pass','ut', 'ksc_azm', 'ksc_elv', 'ksc_dis', 'ksc_ang', 'syo_azm', 'syo_elv', 'syo_dis', 'syo_ang',
            'pra_azm', 'pra_elv', 'pra_dis', 'pra_ang', 'esr_azm', 'esr_elv', 'esr_dis', 'esr_ang', 'gclat','gclon',
            'inv', 'fmlat', 'mlat', 'mlt', 'bmdl_x', 'bmdl_y', 'bmdl_z', 'xxlon_sc', 'xxlat_sc', 'aheight','lsun',
            's_direc_x', 's_direc_y', 's_direc_z', 'sc_pos_x', 'sc_pos_y', 'sc_pos_z', 'sc_vel_x', 'sc_vel_y', 'sc_vel_z']

    try:
        data = load_csv_file(files, cols=cols)
    except UnicodeDecodeError:
        data = load_csv_file(files, cols=cols, gz=True)
    values = data.to_numpy()
    unix_times = time_double([date[2:4] + '-' + date[4:6] + '-' + date[0:2] + '/' + date[6:8] + ':' + date[8:10] + ':' + date[10:12] for date in data['ut']])

    km_in_re = 6374.4

    xyz = np.array([[data['sc_pos_x']], [data['sc_pos_y']], [data['sc_pos_z']]]).transpose([2, 0, 1]).squeeze()
    xyz = np.float64(xyz)
    xyz_re = xyz/km_in_re
    r_theta_phi = xyz_to_polar(xyz)
    rr = r_theta_phi[:, 0]
    th = r_theta_phi[:, 1]
    ph = r_theta_phi[:, 2]
    store_data(prefix + 'geo' + suffix, data={'x': unix_times, 'y': xyz_re})
    store_data(prefix + 'gdlat' + suffix, data={'x': unix_times, 'y': np.float64(data['gclat'])})
    store_data(prefix + 'gdlon' + suffix, data={'x': unix_times, 'y': np.float64(data['gclon'])})
    store_data(prefix + 'inv' +  suffix, data={'x': unix_times, 'y': np.float64(data['inv'])})
    store_data(prefix + 'fmlat' + suffix, data={'x': unix_times, 'y': np.float64(data['fmlat'])})
    store_data(prefix + 'MLT' + suffix, data={'x': unix_times, 'y': np.float64(data['mlt'])})
    store_data(prefix + 'gcalt' + suffix, data={'x': unix_times, 'y': rr / km_in_re})
    store_data(prefix + 'gclat' + suffix, data={'x': unix_times, 'y': th})
    store_data(prefix + 'gclon' + suffix, data={'x': unix_times, 'y': ph})
    options(prefix + 'geo' + suffix, 'ytitle', 'GEO')
    options(prefix + 'geo' + suffix, 'ysubtitle', '[Re]')
    options(prefix + 'gdlat' + suffix, 'ytitle', 'Geodetic latitude of the magnetic footprint')
    options(prefix + 'gdlat' + suffix, 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'gdlon' + suffix, 'ytitle', 'Geodetic longitude of the magnetic footprint')
    options(prefix + 'gdlon' + suffix, 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'inv' + suffix, 'ytitle', 'Invariant Latitude of the magnetic footprint')
    options(prefix + 'inv' + suffix, 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'fmlat' + suffix , 'ytitle', 'Geomagnetic Latitude of the magnetic footprint')
    options(prefix + 'fmlat' + suffix, 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'MLT' + suffix, 'ytitle', 'Magnetic Local Time')
    options(prefix + 'MLT' + suffix, 'ysubtitle', '[hours]')
    options(prefix + 'gcalt' + suffix, 'ytitle', 'Geocentric Altitude')
    options(prefix + 'gcalt' + suffix, 'ysubtitle', '[Re]')
    options(prefix + 'gclat' + suffix, 'ytitle', 'Geocentric Latitude')
    options(prefix + 'gclat' + suffix, 'ysubtitle', '[deg]')
    options(prefix + 'gclon' + suffix, 'ytitle', 'Geocentric Longitude')
    options(prefix + 'gclon' + suffix, 'ysubtitle', '[deg]')

    return [prefix + 'geo' + suffix,
            prefix + 'gdlat' + suffix,
            prefix + 'gdlon' + suffix,
            prefix + 'inv' + suffix,
            prefix + 'fmlat' + suffix,
            prefix + 'MLT' + suffix,
            prefix + 'gcalt' + suffix,
            prefix + 'gclat' + suffix,
            prefix + 'gclon' + suffix]

        




# This routine was originally in akebono/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def pws_postprocessing(variables, prefix='', suffix=''):
    """
    Placeholder for PWS post-processing 
    """
    return variables





import logging
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from .config import CONFIG


def load(
    trange=["2018-10-16", "2018-10-17"],
    instrument="mag",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load DSCOVR data into tplot variables.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ['2018-10-16', '2018-10-17'].
    instrument : str or list of str, optional
        Type of instrument.
        Values can be: 'mag', 'fc' or 'faraday_cup', 'orbit' or 'orb' or 'pre_or', 'att' or 'def_at', 'pre_at', 'all'.
        If 'all' is specified, all instruments will be loaded.
        Default is 'mag'.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
        In all cases a suitable prefix will be given depending on the instrument.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False; only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> mag_vars = pyspedas.projects.dscovr.mag(trange=['2018-11-5', '2018-11-6'])
    >>> tplot('dsc_h0_mag_B1GSE')
    """

    out_files = []
    out_vars = []

    if (
        trange is None
        or not isinstance(trange, list)
        or len(trange) != 2
        or trange[0] > trange[1]
    ):
        logging.error("Invalid trange specified.")
        return out_vars

    if not isinstance(instrument, list):
        instrument = [instrument]
    if "all" in instrument:
        instrument = ["mag", "fc", "orb", "def_at", "pre_at", "pre_or"]

    for instr in instrument:
        if instr == "mag":
            pathformat = "h0/mag/%Y/dscovr_h0_mag_%Y%m%d_v??.cdf"
            pre = "dsc_h0_mag_"
        elif instr == "faraday_cup" or instr == "fc":
            pathformat = "h1/faraday_cup/%Y/dscovr_h1_fc_%Y%m%d_v??.cdf"
            pre = "dsc_h1_fc_"
        elif instr == "def_at" or instr == "att":
            pathformat = "orbit/def_at/%Y/dscovr_at_def_%Y%m%d_v??.cdf"
            pre = "dsc_att_"
        elif instr == "pre_at":
            # Only for year: 2015, 2016, 2017
            pathformat = "orbit/pre_at/%Y/dscovr_at_pre_%Y%m%d_v??.cdf"
            pre = "dsc_at_pre_"
        elif instr in ["pre_or", "orb", "orbit"]:
            pathformat = "orbit/pre_or/%Y/dscovr_orbit_pre_%Y%m%d_v??.cdf"
            pre = "dsc_orbit_"
        else:
            logging.error("Invalid instrument type: " + instr)
            continue

        # If prefix is not empty, add it to the pre variable
        if prefix != "":
            pre = prefix + pre

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        # download the files
        files = download(
            remote_file=remote_names,
            remote_path=CONFIG["remote_data_dir"],
            local_path=CONFIG["local_data_dir"],
            no_download=no_update,
            force_download=force_download,
        )
        if files is not None:
            if not isinstance(files, list):
                files = [files]
            out_files.extend(files)

        if not downloadonly and len(files) > 0:
            # Read the files into tplot variables
            vars = cdf_to_tplot(
                files,
                prefix=pre,
                suffix=suffix,
                get_support_data=get_support_data,
                varformat=varformat,
                varnames=varnames,
                notplot=notplot,
            )
            if not isinstance(vars, list):
                vars = [vars]
            out_vars.extend(vars)

    out_files = list(set(out_files))
    out_files = sorted(out_files)

    if not downloadonly and len(out_files) < 1:
        logging.info("No files were downloaded.")
        return out_files  # return an empty list

    if downloadonly:
        return out_files

    if notplot:
        if len(out_vars) < 1:
            logging.info("No variables were loaded.")
            return {}
        else:
            return out_vars[0]  # return data in hash tables

    if time_clip:
        tclip(out_vars, trange[0], trange[1], suffix="", overwrite=True)

    return out_vars



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot
import numpy as np
import pyspedas
from collections import namedtuple
import logging
from astropy import units as u


def get_data(name, xarray=False, metadata=False, dt=False, units=False, data_quant_in=None, ensure_writeable=False):
    """
    This function extracts the data from the tplot Variables stored in memory.
    
    Parameters
    ----------
        name : str 
            Name of the tplot variable
        xarray : bool, optional
            Keep the variable as an xarray object
        metadata : bool, optional
            Return the metadata of the object (the attr dictionary) instead of the actual data
        dt: bool, optional
            Return the times as np.datetime64[ns] objects instead of unix times
            (significantly faster)
        units: bool, optional
            Attach the astropy units to the data and dependencies prior to returning
        ensure_writeable: bool, optional
            Ensure that returned arrays are writeable, rather than (for example) read-only views of pandas
            data frame indices. Specify 'True' if you need to modify the returned arrays. Defaults to 'False'
            for efficiency.
         
    Returns
    --------
    times: ndarray[float]
        numpy array of seconds since 1970
    y: ndarray
        n-dimensional numpy array of the data values
    v: ndarray
        If exists, an array of bin values for 1-D data types
    spec_bins: ndarray
        If exists, an array of the spectrogram bins for the bin values
    v1: ndarray
        If exists, numpy array of the v1 dimension coordinates
    v2: ndarray
        If exists, numpy array of the v2 dimension coordinates
    v3: ndarray
        If exists, numpy array of the v3 dimension coordinates



    Notes:
    ------

    If metadata==True, the return value is a dict containing variable attributes and plot options.
    Otherwise, the return value is a named tuple with the fields described above.

    Examples
    --------
        >>> # Retrieve the data from Variable 1
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> time, data = pyspedas.get_data("Variable1")
        >>> metadata = pyspedas.get_data("Variable1", metadata=True)

    """

    #check for an input data_quant object
    if data_quant_in is not None:
        temp_data_quant = data_quant_in
    else:
        if name not in pyspedas.tplot_tools.data_quants.keys():
            logging.info("The name " + str(name) + " is currently not in pyspedas")
            return
    
        temp_data_quant = pyspedas.tplot_tools.data_quants[name]

    if isinstance(temp_data_quant, dict):
        # non-record varying variables are stored as dicts
        return temp_data_quant['data']

    if xarray:
        return temp_data_quant

    if metadata:
        return temp_data_quant.attrs

    error = temp_data_quant.attrs['plot_options']['error']

    if not dt:
        # TODO: Is this always at ns resolution?  ERG LEPE CDFs seem to be in microseconds
        times = np.int64(temp_data_quant.time.values)/1e9
        #times = np.array([int(time)/1e9 for time in temp_data_quant.time.values])
        #times = np.array([(time - np.datetime64('1970-01-01T00:00:00'))/np.timedelta64(1, 's') for time in temp_data_quant.time.values])
    else:
        times = temp_data_quant.time.values

    coord_names = temp_data_quant.coords.keys()
    data_values = temp_data_quant.data

    # Temporary patch for Pandas 3.0.0
    # Arguably, this is dangerous and should not be allowed...
    if ensure_writeable and not times.flags['WRITEABLE']:
        times=times.copy()
    if ensure_writeable and not data_values.flags['WRITEABLE']:
        data_values=data_values.copy()

    v1_values = None
    v2_values = None
    v3_values = None

    data_att_set = 'data_att' in temp_data_quant.attrs

    if 'v' in coord_names:
        v1_values = temp_data_quant.coords['v'].values
    if 'spec_bins' in coord_names:
        v1_values = temp_data_quant.coords['spec_bins'].values
    if 'v1' in coord_names:
        v1_values = temp_data_quant.coords['v1'].values
    if 'v2' in coord_names:
        v2_values = temp_data_quant.coords['v2'].values
    if 'v3' in coord_names:
        v3_values = temp_data_quant.coords['v3'].values

    if data_att_set and units:
        data_units = temp_data_quant.attrs['data_att'].get('units')
        v1_units = temp_data_quant.attrs['data_att'].get('depend_1_units')
        v2_units = temp_data_quant.attrs['data_att'].get('depend_2_units')
        v3_units = temp_data_quant.attrs['data_att'].get('depend_3_units')

        try:
            if data_units is not None:
                data_values = data_values * u.Unit(data_units)
            if v1_values is not None and v1_units is not None:
                v1_values = v1_values * u.Unit(v1_units)
            if v2_values is not None and v2_units is not None:
                v2_values = v2_values * u.Unit(v2_units)
            if v3_values is not None and v3_units is not None:
                v3_values = v3_values * u.Unit(v3_units)
        except ValueError:
            # occurs when there's a problem converting the units string
            # to astropy units
            pass

    if 'v1' in coord_names and 'v2' in coord_names and 'v3' in coord_names:
        if ensure_writeable and not v1_values.flags['WRITEABLE']:
            v1_values = v1_values.copy()
        if ensure_writeable and not v2_values.flags['WRITEABLE']:
            v2_values = v2_values.copy()
        if ensure_writeable and not v3_values.flags['WRITEABLE']:
            v3_values = v3_values.copy()
        variable = namedtuple('variable', ['times', 'y', 'v1', 'v2', 'v3'])
        return variable(times, data_values, v1_values, v2_values, v3_values)
    elif 'v1' in coord_names and 'v2' in coord_names:
        if ensure_writeable and not v1_values.flags['WRITEABLE']:
            v1_values = v1_values.copy()
        if ensure_writeable and not v2_values.flags['WRITEABLE']:
            v2_values = v2_values.copy()

        variable = namedtuple('variable', ['times', 'y', 'v1', 'v2'])
        return variable(times, data_values, v1_values, v2_values)
    elif 'v1' in coord_names:
        if ensure_writeable and not v1_values.flags['WRITEABLE']:
            v1_values = v1_values.copy()

        variable = namedtuple('variable', ['times', 'y', 'v1'])
        return variable(times, data_values, v1_values)
    elif 'v' in coord_names:
        if ensure_writeable and not v1_values.flags['WRITEABLE']:
            v1_values = v1_values.copy()

        variable = namedtuple('variable', ['times', 'y', 'v'])
        return variable(times, data_values, v1_values)
    elif 'spec_bins' in coord_names:
        if ensure_writeable and not v1_values.flags['WRITEABLE']:
            v1_values = v1_values.copy()

        variable = namedtuple('variable', ['times', 'y', 'v'])
        return variable(times, data_values, v1_values)

    if error is not None:
        if ensure_writeable and not error.flags['WRITEABLE']:
            error = error.copy()
        variable = namedtuple('variable', ['times', 'y', 'dy'])
        return variable(times, data_values, error)
    else:
        variable = namedtuple('variable', ['times', 'y'])
        return variable(times, data_values)


def get(name, xarray=False, metadata=False, dt=True, units=True, ensure_writeable=False):
    """
    This function extracts the data from the tplot Variables stored in memory.

    Parameters:
        name : str
            Name of the tplot variable
        xarray : bool
            Keep the variable as an xarray object
        metadata : bool
            Return the metadata of the object (the attr dictionary) instead of the actual data
        dt : bool
            Return the times as np.datetime64[ns] objects instead of unix times
            (significantly faster); defaults to True for pyspedas.get
        units: bool
            Attach the astropy units to the data and dependencies prior to returning
            defaults to True for pyspedas.get
        ensure_writeable: bool, optional
            Ensure that returned arrays are writeable, rather than (for example) read-only views of pandas
            data frame indices. Specify 'True' if you need to modify the returned arrays. Defaults to 'False'
            for efficiency.


    Returns: tuple of data/dimensions/metadata stored in pyspedas
        time_val : numpy array of seconds since 1970
        data_val : n-dimensional array of data
        spec_bins_val (if exists) : spectral bins if the plot is a spectrogram
        v1_val (if exists) : numpy array of v1 dimension coordinates
        v2_val {if exists} : numpy array of v2 dimension coordinates
        v3_val (if exists) : numpy array of v3 dimension coordinates


    Examples:
        >>> # Retrieve the data from Variable 1
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store("Variable1", data={'x':x_data, 'y':y_data})
        >>> time, data = pyspedas.get("Variable1")

    """
    return get_data(name, xarray=xarray, metadata=metadata, dt=dt, units=units, ensure_writeable=ensure_writeable)




import logging
from pyspedas import wildcard_expand
from pyspedas.projects.themis.load import load


def efi(trange=['2007-03-23', '2007-03-24'],
        probe='c',
        level='l2',
        datatype=None,
        suffix='',
        get_support_data=False,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads Electric Field Instrument (EFI) data

    Parameters
    ----------

        trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2007-03-23', '2007-03-24']

        probe: str or list of str
            Spacecraft probe letter(s) ('a', 'b', 'c', 'd' and/or 'e')
            Default: 'c'

        level: str
            Processing level; Valid options: 'l1', 'l2'
            Default: 'l2'

        datatype: str or list of str
            Data type; Valid L1 options::

                'eff', Fast survey E12, E34, E56 waveforms
                'efp', Particle burst E12, E34, E56 waveforms
                'efw', Wave burst E12 E34, E56 waveforms
                'vaf', Fast survey voltage group A, V1-V6 boom voltages
                'vap', Particle burst voltage group A, V1-V6 boom voltages
                'vaw', Wave burst voltage group A, V1-V6 boom voltages
                'vbf', Fast survey voltage group B, V1-V6 boom voltages
                'vbp', Particle burst voltage group B, V1-V6 boom voltages
                'vbw', Wave burst voltage group B, V1-V6 boom voltages
                L1 default: [eff. efp, efw, vaf. vap, vaw]

            Valid L2 options::

                'efi', Fast survey E field vectors and other quantities
                'efp', Particle burst E field vectors
                'efw', Wave burst E field vectors
                L2 default: efi

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  
            Default: False; only loads data with a "VAR_TYPE" attribute of "data"

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None; all variables are loaded

        varnames: list of str
            List of variable names to load
            Default: Empty list, so all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword
            Default: False

    Returns
    -------
    List of str
        List of tplot variables created
        Empty list if no data
    
    Example
    -------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> efi_vars = pyspedas.projects.themis.efi(probe='d', trange=['2013-11-5', '2013-11-6'])
        >>> tplot('thd_efs_dot0_gse')


    """
    valid_levels = ['l1', 'l2']
    valid_l1_datatypes = ['eff', 'efp', 'efw', 'vaf', 'vap', 'vaw', 'vbf', 'vbp', 'vbw']
    default_l1_datatypes = ['eff', 'efp', 'efw', 'vaf', 'vap', 'vaw']  # omit vb* by default
    valid_l2_datatypes = ['efi', 'efp', 'efw']
    default_l2_datatypes = ['efi'] # omit efp and efw unless specifically requested

    if level.lower() not in valid_levels:
        logging.error("Unrecognized level %s", level)
        return []

    level=level.lower()

    if level == 'l1':
        valid_datatypes=valid_l1_datatypes
        default_datatypes=default_l1_datatypes
    else:
        valid_datatypes=valid_l2_datatypes
        default_datatypes=default_l2_datatypes

    if datatype is None:
        selected_datatype=default_datatypes
    else:
        selected_datatype=wildcard_expand(valid_datatypes, datatype, case_sensitive=False)

    if len(selected_datatype) == 0:
        logging.error("No valid datatypes selected")
        return []

    return load(instrument='efi', trange=trange, level=level,
                datatype=selected_datatype,
                suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, varnames=varnames,
                downloadonly=downloadonly, notplot=notplot,
                probe=probe, time_clip=time_clip, no_update=no_update)



import os
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_none_data(self):
        induction_vars = pyspedas.projects.mica.induction()
        self.assertTrue(induction_vars == None)

    def test_load_notplot(self):
        induction_vars = pyspedas.projects.mica.induction(site='nal', notplot=True)
        self.assertTrue('spectra_x_1Hz_NAL' in induction_vars)

    def test_load_NAL_data(self):
        induction_vars = pyspedas.projects.mica.induction(site='nal', time_clip=True)
        self.assertTrue(data_exists('spectra_x_1Hz_NAL'))

    def test_downloadonly(self):
        files = pyspedas.projects.mica.induction(site='nal', downloadonly=True, trange=['2014-2-15', '2014-2-16'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()
